#!/usr/bin/env python3
# coding: utf-8
"""
Created on Jun 2020
@author: Victor Manach <victor.manach@telecom-paris.fr>
@author: Rémi Jaylet <remi.jaylet@telecom-paris.fr>
"""
from typing import Optional, Union

import numpy as np
from scipy import sparse
from scipy.spatial import cKDTree

from sknetwork.embedding.base import BaseEmbedding
from sknetwork.utils.check import check_format, is_symmetric, check_square
from sknetwork.utils.format import directed2undirected


class ForceAtlas(BaseEmbedding):
    """Force Atlas layout for displaying graphs.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_components : int
        Dimension of the graph layout.
    n_iter : int
        Number of iterations to update positions.
        If ``None``, use the value of self.n_iter.
    approx_radius : float
        If a positive value is provided, only the nodes within this distance a given node are used to compute
        the repulsive force.
    lin_log : bool
        If ``True``, use lin-log mode.
    gravity_factor : float
        Gravity force scaling constant.
    repulsive_factor : float
        Repulsive force scaling constant.
    tolerance : float
        Tolerance defined in the swinging constant.
    speed : float
        Speed constant.
    speed_max : float
        Constant used to impose constrain on speed.

    Attributes
    ----------
    embedding_ : np.ndarray
        Layout in given dimension.

    Example
    -------
    >>> from sknetwork.embedding.force_atlas import ForceAtlas
    >>> from sknetwork.data import karate_club
    >>> force_atlas = ForceAtlas()
    >>> adjacency = karate_club()
    >>> embedding = force_atlas.fit_transform(adjacency)
    >>> embedding.shape
    (34, 2)

    References
    ----------
    Jacomy M., Venturini T., Heymann S., Bastian M. (2014).
    `ForceAtlas2, a Continuous Graph Layout Algorithm for Handy Network Visualization Designed for the Gephi Software.
    <https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0098679>`_
    Plos One.
    """
    def __init__(self, n_components: int = 2, n_iter: int = 50, approx_radius: float = -1, lin_log: bool = False,
                 gravity_factor: float = 0.01, repulsive_factor: float = 0.1, tolerance: float = 0.1,
                 speed: float = 0.1, speed_max: float = 10):
        super(ForceAtlas, self).__init__()
        self.n_components = n_components
        self.n_iter = n_iter
        self.approx_radius = approx_radius
        self.lin_log = lin_log
        self.gravity_factor = gravity_factor
        self.repulsive_factor = repulsive_factor
        self.tolerance = tolerance
        self.speed = speed
        self.speed_max = speed_max

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray], pos_init: Optional[np.ndarray] = None,
            n_iter: Optional[int] = None) -> 'ForceAtlas':
        """Compute layout.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph, treated as undirected.
        pos_init :
            Position to start with. Random if not provided.
        n_iter : int
            Number of iterations to update positions.
            If ``None``, use the value of self.n_iter.

        Returns
        -------
        self: :class:`ForceAtlas`
        """
        # verify the format of the adjacency matrix
        adjacency = check_format(adjacency)
        check_square(adjacency)
        if not is_symmetric(adjacency):
            adjacency = directed2undirected(adjacency)
        n = adjacency.shape[0]

        # setting of the tolerance according to the size of the graph
        if n < 5000:
            tolerance = 0.1
        elif 5000 <= n < 50000:  # pragma: no cover
            tolerance = 1
        else:  # pragma: no cover
            tolerance = 10

        if n_iter is None:
            n_iter = self.n_iter

        # initial position of the nodes of the graph
        if pos_init is None:
            position: np.ndarray = np.random.randn(n, self.n_components)
        else:
            if pos_init.shape != (n, self.n_components):
                raise ValueError('The initial position does not have valid dimensions.')
            else:
                position = pos_init
        # compute the vector with the degree of each node
        degree: np.ndarray = adjacency.dot(np.ones(adjacency.shape[1])) + 1

        # initialization of variation of position of nodes
        resultants = np.zeros(n)
        delta: np.ndarray = np.zeros((n, self.n_components))
        swing_vector: np.ndarray = np.zeros(n)
        global_speed = 1

        for iteration in range(n_iter):
            delta *= 0
            global_swing = 0
            global_traction = 0

            if self.approx_radius > 0:
                tree = cKDTree(position)
            else:
                tree = None

            for i in range(n):

                # attraction
                indices = adjacency.indices[adjacency.indptr[i]:adjacency.indptr[i + 1]]
                attraction = position[i] - position[indices]

                if self.lin_log:
                    attraction = np.sign(attraction) * np.log(1 + np.abs(10 * attraction))
                attraction = attraction.sum(axis=0)

                # repulsion
                if tree is None:
                    neighbors = np.arange(n)
                else:
                    neighbors = tree.query_ball_point(position[i], self.approx_radius)

                grad: np.ndarray = (position[i] - position[neighbors])  # shape (n_neigh, n_components)
                distance: np.ndarray = np.linalg.norm(grad, axis=1)  # shape (n_neigh,)
                distance = np.where(distance < 0.01, 0.01, distance)
                repulsion = grad * (degree[neighbors] / distance)[:, np.newaxis]

                repulsion *= self.repulsive_factor * degree[i]
                repulsion = repulsion.sum(axis=0)

                # gravity
                gravity = self.gravity_factor * degree[i] * grad
                gravity = gravity.sum(axis=0)

                # forces resultant applied on node i for traction, swing and speed computation
                force = repulsion - attraction - gravity
                resultant_new: float = np.linalg.norm(force)
                resultant_old: float = resultants[i]

                swing_node: float = np.abs(resultant_new - resultant_old)  # force variation applied on node i
                swing_vector[i] = swing_node
                global_swing += (degree[i] + 1) * swing_node

                traction: float = np.abs(resultant_new + resultant_old) / 2  # traction force applied on node i
                global_traction += (degree[i] + 1) * traction

                node_speed = self.speed * global_speed / (1 + global_speed * np.sqrt(swing_node))
                if node_speed > self.speed_max / resultant_new:  # pragma: no cover
                    node_speed = self.speed_max / resultant_new

                delta[i]: np.ndarray = node_speed * force
                resultants[i] = resultant_new
                global_speed = tolerance * global_traction / global_swing

            position += delta  # calculating displacement and final position of points after iteration
            if (swing_vector < 1).all():
                break  # if the swing of all nodes is zero, then convergence is reached and we break.

        self.embedding_ = position
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Jul 24, 2019
"""
import numpy as np
from scipy import sparse

from sknetwork.utils.check import is_symmetric


def breadth_first_search(adjacency: sparse.csr_matrix, source: int, return_predecessors: bool = True):
    """Breadth-first ordering starting with specified node.

    * Graphs
    * Digraphs

    Based on SciPy (scipy.sparse.csgraph.breadth_first_order)

    Parameters
    ----------
    adjacency :
        The adjacency matrix of the graph
    source : int
        The node from which to start the ordering
    return_predecessors : bool
        If ``True``, the size predecessor matrix is returned

    Returns
    -------
    node_array : np.ndarray
        The breadth-first list of nodes, starting with specified node. The length of node_array is the number of nodes
        reachable from the specified node.
    predecessors : np.ndarray
        Returned only if ``return_predecessors == True``. The list of predecessors of each node in a breadth-first tree.
        If node ``i`` is in the tree, then its parent is given by ``predecessors[i]``. If node ``i`` is not in the tree
        (and for the parent node) then ``predecessors[i] = -9999``.
    """
    directed = not is_symmetric(adjacency)
    return sparse.csgraph.breadth_first_order(adjacency, source, directed, return_predecessors)


def depth_first_search(adjacency: sparse.csr_matrix, source: int, return_predecessors: bool = True):
    """Depth-first ordering starting with specified node.

    * Graphs
    * Digraphs

    Based on SciPy (scipy.sparse.csgraph.depth_first_order)

    Parameters
    ----------
    adjacency :
        The adjacency matrix of the graph
    source :
        The node from which to start the ordering
    return_predecessors:
        If ``True``, the size predecessor matrix is returned

    Returns
    -------
    node_array : np.ndarray
        The depth-first list of nodes, starting with specified node. The length of node_array is the number of nodes
        reachable from the specified node.
    predecessors : np.ndarray
        Returned only if ``return_predecessors == True``. The list of predecessors of each node in a depth-first tree.
        If node ``i`` is in the tree, then its parent is given by ``predecessors[i]``. If node ``i`` is not in the tree
        (and for the parent node) then ``predecessors[i] = -9999``.
    """
    directed = not is_symmetric(adjacency)
    return sparse.csgraph.depth_first_order(adjacency, source, directed, return_predecessors)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Louvain"""
import unittest

import numpy as np

from sknetwork.data import house, edgelist2adjacency
from sknetwork.linalg.operators import RegularizedAdjacency
from sknetwork.linalg.ppr_solver import get_pagerank
from sknetwork.utils.check import is_proba_array


class TestPPR(unittest.TestCase):

    def test_diteration(self):
        # test convergence by tolerance
        adjacency = house()
        seeds = np.ones(adjacency.shape[0]) / adjacency.shape[0]
        pr = get_pagerank(adjacency, damping_factor=0.85, n_iter=100, tol=10, solver='diteration', seeds=seeds)
        self.assertTrue(is_proba_array(pr))

        # test graph with some null out-degree
        adjacency = edgelist2adjacency([(0, 1)])
        seeds = np.ones(adjacency.shape[0]) / adjacency.shape[0]
        pr = get_pagerank(adjacency, damping_factor=0.85, n_iter=100, tol=10, solver='diteration', seeds=seeds)
        self.assertTrue(is_proba_array(pr))

        # test invalid entry
        adjacency = RegularizedAdjacency(house(), 0.1)
        seeds = np.ones(adjacency.shape[0]) / adjacency.shape[0]
        with self.assertRaises(ValueError):
            get_pagerank(adjacency, damping_factor=0.85, n_iter=100, tol=10, solver='diteration', seeds=seeds)

    def test_piteration(self):
        # test on SparseLR matrix
        adjacency = RegularizedAdjacency(house(), 0.1)
        seeds = np.ones(adjacency.shape[0]) / adjacency.shape[0]
        pr = get_pagerank(adjacency, damping_factor=0.85, n_iter=100, tol=10, solver='piteration', seeds=seeds)
        self.assertTrue(is_proba_array(pr))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov 6 2018
@author: Nathan De Lara <ndelara@enst.fr>
Quality metrics for adjacency embeddings
"""
import numpy as np
from scipy import sparse

from sknetwork.linalg import normalize
from sknetwork.utils.check import check_format, check_probs, check_square


def cosine_modularity(adjacency, embedding: np.ndarray, embedding_col=None, resolution=1., weights='degree',
                      return_all: bool = False):
    """Quality metric of an embedding :math:`x` defined by:

    :math:`Q = \\sum_{ij}\\left(\\dfrac{A_{ij}}{w} - \\gamma \\dfrac{w^+_iw^-_j}{w^2}\\right)
    \\left(\\dfrac{1 + \\cos(x_i, x_j)}{2}\\right)`

    where

    * :math:`w^+_i, w^-_i` are the out-weight, in-weight of node :math:`i` (for digraphs),\n
    * :math:`w = 1^TA1` is the total weight of the graph.

    For bipartite graphs with column embedding :math:`y`, the metric is

    :math:`Q = \\sum_{ij}\\left(\\dfrac{B_{ij}}{w} - \\gamma \\dfrac{w_{1,i}w_{2,j}}{w^2}\\right)
    \\left(\\dfrac{1 + \\cos(x_i, y_j)}{2}\\right)`

    where

    * :math:`w_{1,i}, w_{2,j}` are the weights of nodes :math:`i` (row) and :math:`j` (column),\n
    * :math:`w = 1^TB1` is the total weight of the graph.

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph.
    embedding :
        Embedding of the nodes.
    embedding_col :
        Embedding of the columns (for bipartite graphs).
    resolution :
        Resolution parameter.
    weights : ``'degree'`` or ``'uniform'``
        Weights of the nodes.
    return_all :
        If ``True``, also return fit and diversity

    Returns
    -------
    modularity : float
    fit: float, optional
    diversity: float, optional

    Example
    -------
    >>> from sknetwork.embedding import cosine_modularity
    >>> from sknetwork.data import karate_club
    >>> graph = karate_club(metadata=True)
    >>> adjacency = graph.adjacency
    >>> embedding = graph.position
    >>> np.round(cosine_modularity(adjacency, embedding), 2)
    0.35
    """
    adjacency = check_format(adjacency)
    total_weight: float = adjacency.data.sum()

    if embedding_col is None:
        check_square(adjacency)
        embedding_col = embedding.copy()

    embedding_row_norm = normalize(embedding, p=2)
    embedding_col_norm = normalize(embedding_col, p=2)

    probs_row = check_probs(weights, adjacency)
    probs_col = check_probs(weights, adjacency.T)

    if isinstance(embedding_row_norm, sparse.csr_matrix) and isinstance(embedding_col_norm, sparse.csr_matrix):
        fit: float = 0.5 * (1 + (embedding_row_norm.multiply(adjacency.dot(embedding_col_norm))).sum() / total_weight)
    else:
        fit: float = 0.5 * (
            1 + (np.multiply(embedding_row_norm, adjacency.dot(embedding_col_norm))).sum() / total_weight)
    div: float = 0.5 * (1 + (embedding.T.dot(probs_row)).dot(embedding_col.T.dot(probs_col)))

    if return_all:
        return fit, div, fit - resolution * div
    else:
        return fit - resolution * div



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on November 15, 2019
@author: Quentin Lutz <qlutz@enst.fr>
"""

import pickle
import shutil
import tarfile
from os import environ, makedirs, remove, listdir, rmdir
from os.path import exists, expanduser, join
from pathlib import Path
from typing import Optional, Union
from urllib.error import HTTPError, URLError
from urllib.request import urlretrieve

import numpy as np
from scipy import sparse

from sknetwork.data.parse import load_edge_list, load_labels, load_header, load_metadata
from sknetwork.utils import Bunch
from sknetwork.utils.check import is_square

NETSET_URL = 'https://netset.telecom-paris.fr'


def get_data_home(data_home: Optional[Union[str, Path]] = None) -> Path:
    """
    Return a path to a storage folder depending on the dedicated environment variable and user input.

    Parameters
    ----------
    data_home: str
        The folder to be used for dataset storage
    """
    if data_home is None:
        data_home = environ.get('SCIKIT_NETWORK_DATA',
                                join('~', 'scikit_network_data'))
    data_home = expanduser(data_home)
    if not exists(data_home):
        makedirs(data_home)
    return Path(data_home)


def clear_data_home(data_home: Optional[Union[str, Path]] = None):
    """
    Clear storage folder depending on the dedicated environment variable and user input.

    Parameters
    ----------
    data_home: str or :class:`pathlib.Path`
        The folder to be used for dataset storage
    """
    data_home = get_data_home(data_home)
    shutil.rmtree(data_home)


def load_netset(dataset: Optional[str] = None, data_home: Optional[Union[str, Path]] = None) -> Bunch:
    """Load a dataset from the `NetSet database
    <https://netset.telecom-paris.fr/>`_.

    Parameters
    ----------
    dataset : str
        The name of the dataset (all low-case). Examples include 'openflights', 'cinema' and 'wikivitals'.
    data_home : str or :class:`pathlib.Path`
        The folder to be used for dataset storage.

    Returns
    -------
    graph : :class:`Bunch`
    """
    graph = Bunch()
    npz_folder = NETSET_URL + '/datasets_npz/'

    if dataset is None:
        print("Please specify the dataset (e.g., 'openflights' or 'wikivitals').\n" +
              f"Complete list available here: <{npz_folder}>.")
        return graph
    else:
        dataset = dataset.lower()
    data_home = get_data_home(data_home)
    data_path = data_home / dataset
    if not data_path.exists():
        makedirs(data_path, exist_ok=True)
        try:
            urlretrieve(npz_folder + dataset + '_npz.tar.gz',
                        data_home / (dataset + '_npz.tar.gz'))
        except HTTPError:
            rmdir(data_path)
            raise ValueError('Invalid dataset: ' + dataset + '.'
                             + "\nAvailable datasets include 'openflights' and 'wikivitals'."
                             + f"\nSee <{NETSET_URL}>")
        except ConnectionResetError:  # pragma: no cover
            rmdir(data_path)
            raise RuntimeError("Could not reach Netset.")
        with tarfile.open(data_home / (dataset + '_npz.tar.gz'), 'r:gz') as tar_ref:
            tar_ref.extractall(data_home)
        remove(data_home / (dataset + '_npz.tar.gz'))

    files = [file for file in listdir(data_path)]

    for file in files:
        file_components = file.split('.')
        if len(file_components) == 2:
            file_name, file_extension = tuple(file_components)
            if file_extension == 'npz':
                graph[file_name] = sparse.load_npz(data_path / file)
            elif file_extension == 'npy':
                graph[file_name] = np.load(data_path / file)
            elif file_extension == 'p':
                with open(data_path / file, 'rb') as f:
                    graph[file_name] = pickle.load(f)

    return graph


def load_konect(dataset: str, data_home: Optional[Union[str, Path]] = None, auto_numpy_bundle: bool = True) -> Bunch:
    """Load a dataset from the `Konect database
    <http://konect.cc/networks/>`_.

    Parameters
    ----------
    dataset : str
        The internal name of the dataset as specified on the Konect website (e.g. for the Zachary Karate club dataset,
        the corresponding name is ``'ucidata-zachary'``).
    data_home : str or :class:`pathlib.Path`
        The folder to be used for dataset storage
    auto_numpy_bundle : bool
        Denotes if the dataset should be stored in its default format (False) or using Numpy files for faster
        subsequent access to the dataset (True).

    Returns
    -------
    graph : :class:`Bunch`
        An object with the following attributes:

             * `adjacency` or `biadjacency`: the adjacency/biadjacency matrix for the dataset
             * `meta`: a dictionary containing the metadata as specified by Konect
             * each attribute specified by Konect (ent.* file)

    Notes
    -----
    An attribute `meta` of the `Bunch` class is used to store information about the dataset if present. In any case,
    `meta` has the attribute `name` which, if not given, is equal to the name of the dataset as passed to this function.

    References
    ----------
    Kunegis, J. (2013, May).
    `Konect: the Koblenz network collection.
    <https://dl.acm.org/doi/abs/10.1145/2487788.2488173>`_
    In Proceedings of the 22nd International Conference on World Wide Web (pp. 1343-1350).
    """
    if dataset == '':
        raise ValueError("Please specify the dataset. "
                         + "\nExamples include 'actor-movie' and 'ego-facebook'."
                         + "\n See 'http://konect.cc/networks/' for the full list.")
    data_home = get_data_home(data_home)
    data_path = data_home / dataset
    if not data_path.exists():
        makedirs(data_path, exist_ok=True)
        try:
            urlretrieve('http://konect.cc/files/download.tsv.' + dataset + '.tar.bz2',
                        data_home / (dataset + '.tar.bz2'))
            with tarfile.open(data_home / (dataset + '.tar.bz2'), 'r:bz2') as tar_ref:
                tar_ref.extractall(data_home)
        except (HTTPError, tarfile.ReadError):
            rmdir(data_path)
            raise ValueError('Invalid dataset ' + dataset + '.'
                             + "\nExamples include 'actor-movie' and 'ego-facebook'."
                             + "\n See 'http://konect.cc/networks/' for the full list.")
        except (URLError, ConnectionResetError):  # pragma: no cover
            rmdir(data_path)
            raise RuntimeError("Could not reach Konect.")
        finally:
            if exists(data_home / (dataset + '.tar.bz2')):
                remove(data_home / (dataset + '.tar.bz2'))
    elif exists(data_path / (dataset + '_bundle')):
        return load_from_numpy_bundle(dataset + '_bundle', data_path)

    data = Bunch()

    files = [file for file in listdir(data_path) if dataset in file]

    matrix = [file for file in files if 'out.' in file]
    if matrix:
        file = matrix[0]
        directed, bipartite, weighted = load_header(data_path / file)
        if bipartite:
            graph = load_edge_list(data_path / file, directed=directed, bipartite=bipartite, weighted=weighted)
            data.biadjacency = graph.biadjacency
        else:
            graph = load_edge_list(data_path / file, directed=directed, bipartite=bipartite, weighted=weighted)
            data.adjacency = graph.adjacency

    metadata = [file for file in files if 'meta.' in file]
    if metadata:
        file = metadata[0]
        data.meta = load_metadata(data_path / file)

    attributes = [file for file in files if 'ent.' + dataset in file]
    if attributes:
        for file in attributes:
            attribute_name = file.split('.')[-1]
            data[attribute_name] = load_labels(data_path / file)

    if hasattr(data, 'meta'):
        if hasattr(data.meta, 'name'):
            pass
        else:
            data.meta.name = dataset
    else:
        data.meta = Bunch()
        data.meta.name = dataset

    if auto_numpy_bundle:
        save_to_numpy_bundle(data, dataset + '_bundle', data_path)

    return data


def save_to_numpy_bundle(data: Bunch, bundle_name: str, data_home: Optional[Union[str, Path]] = None):
    """Save a Bunch in the specified data home to a collection of Numpy and Pickle files for faster subsequent loads.

    Parameters
    ----------
    data: Bunch
        The data to save
    bundle_name: str
        The name to be used for the bundle folder
    data_home: str or :class:`pathlib.Path`
        The folder to be used for dataset storage
    """
    data_home = get_data_home(data_home)
    data_path = data_home / bundle_name
    makedirs(data_path, exist_ok=True)
    for attribute in data:
        if type(data[attribute]) == sparse.csr_matrix:
            sparse.save_npz(data_path / attribute, data[attribute])
        elif type(data[attribute]) == np.ndarray:
            np.save(data_path / attribute, data[attribute])
        elif type(data[attribute]) == Bunch or type(data[attribute]) == str:
            with open(data_path / (attribute + '.p'), 'wb') as file:
                pickle.dump(data[attribute], file)
        else:
            raise TypeError('Unsupported data attribute type '+str(type(data[attribute])) + '.')


def load_from_numpy_bundle(bundle_name: str, data_home: Optional[Union[str, Path]] = None):
    """Load a Bunch from a collection of Numpy and Pickle files (inverse function of ``save_to_numpy_bundle``).

    Parameters
    ----------
    bundle_name: str
        The name used for the bundle folder
    data_home: str or :class:`pathlib.Path`
        The folder used for dataset storage

    Returns
    -------
    data: Bunch
        The original data
    """
    data_home = get_data_home(data_home)
    data_path = data_home / bundle_name
    if not data_path.exists():
        raise FileNotFoundError('No bundle at ' + str(data_path))
    else:
        files = listdir(data_path)
        data = Bunch()
        for file in files:
            if len(file.split('.')) == 2:
                file_name, file_extension = file.split('.')
                if file_extension == 'npz':
                    data[file_name] = sparse.load_npz(data_path / file)
                elif file_extension == 'npy':
                    data[file_name] = np.load(data_path / file)
                elif file_extension == 'p':
                    with open(data_path / file, 'rb') as f:
                        data[file_name] = pickle.load(f)
        return data


def save(folder: Union[str, Path], data: Union[sparse.csr_matrix, Bunch]):
    """Save a Bunch or a CSR matrix in the current directory to a collection of Numpy and Pickle files for faster
    subsequent loads. Supported attribute types include sparse matrices, NumPy arrays, strings and Bunch.

    Parameters
    ----------
    folder : str or :class:`pathlib.Path`
        The name to be used for the bundle folder
    data : Union[sparse.csr_matrix, Bunch]
        The data to save

    Example
    -------
    >>> from sknetwork.data import save
    >>> graph = Bunch()
    >>> graph.adjacency = sparse.csr_matrix(np.random.random((10, 10)) < 0.2)
    >>> graph.names = np.array(list('abcdefghij'))
    >>> save('random_data', graph)
    >>> 'random_data' in listdir('.')
    True
    """
    folder = Path(folder)
    folder = folder.expanduser()
    if folder.exists():
        shutil.rmtree(folder)
    if isinstance(data, sparse.csr_matrix):
        bunch = Bunch()
        if is_square(data):
            bunch.adjacency = data
        else:
            bunch.biadjacency = data
        data = bunch
    if folder.is_absolute():
        save_to_numpy_bundle(data, folder, '/')
    else:
        save_to_numpy_bundle(data, folder, '.')


def load(folder: Union[str, Path]):
    """Load a Bunch from a previously created bundle from the current directory (inverse function of ``save``).

    Parameters
    ----------
    folder: str
        The name used for the bundle folder

    Returns
    -------
    data: Bunch
        The original data

    Example
    -------
    >>> from sknetwork.data import save
    >>> graph = Bunch()
    >>> graph.adjacency = sparse.csr_matrix(np.random.random((10, 10)) < 0.2)
    >>> graph.names = np.array(list('abcdefghij'))
    >>> save('random_data', graph)
    >>> loaded_graph = load('random_data')
    >>> loaded_graph.names[0]
    'a'
    """
    folder = Path(folder)
    if folder.is_absolute():
        return load_from_numpy_bundle(folder, '/')
    else:
        return load_from_numpy_bundle(folder, '.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for eig"""

import unittest

import numpy as np

from sknetwork.data import miserables, karate_club
from sknetwork.linalg import LanczosEig, HalkoEig, SparseLR


def eigenvector_err(matrix, eigenvectors, eigenvalues):
    """Approximation error for eigenvectors."""
    err = matrix.dot(eigenvectors) - eigenvectors * eigenvalues
    return np.linalg.norm(err)


# noinspection DuplicatedCode
class TestSolvers(unittest.TestCase):

    def setUp(self):
        """Instanciate les Miserables and regularized version"""
        self.adjacency = miserables()
        self.random_state = np.random.RandomState(123)
        n = self.adjacency.shape[0]
        x = np.random.random(n)
        self.slr = SparseLR(self.adjacency, [(x, x)])

    def test_lanczos(self):
        solver = LanczosEig('LM')
        solver.fit(self.adjacency, 2)
        self.assertEqual(len(solver.eigenvalues_), 2)
        self.assertAlmostEqual(eigenvector_err(self.adjacency, solver.eigenvectors_, solver.eigenvalues_), 0)

        solver.fit(self.slr, 2)
        self.assertEqual(len(solver.eigenvalues_), 2)
        self.assertAlmostEqual(eigenvector_err(self.slr, solver.eigenvectors_, solver.eigenvalues_), 0)

        adjacency = karate_club()
        solver = LanczosEig('SM')
        solver.fit(adjacency, 2)
        self.assertEqual(len(solver.eigenvalues_), 2)
        self.assertAlmostEqual(eigenvector_err(adjacency, solver.eigenvectors_, solver.eigenvalues_), 0)

    def test_halko(self):
        solver = HalkoEig('LM', random_state=self.random_state)
        solver.fit(self.adjacency, 2)
        self.assertEqual(len(solver.eigenvalues_), 2)
        self.assertAlmostEqual(eigenvector_err(self.adjacency, solver.eigenvectors_, solver.eigenvalues_), 0)

        solver.fit(self.slr, 2)
        self.assertEqual(len(solver.eigenvalues_), 2)
        self.assertAlmostEqual(eigenvector_err(self.slr, solver.eigenvectors_, solver.eigenvalues_), 0)

        solver = HalkoEig('SM', random_state=self.random_state)
        solver.fit(self.adjacency, 2)
        self.assertEqual(len(solver.eigenvalues_), 2)
        # self.assertAlmostEqual(eigenvector_err(self.adjacency, solver.eigenvectors_, solver.eigenvalues_), 0)

    def test_compare_solvers(self):
        lanczos = LanczosEig('LM')
        halko = HalkoEig('LM', random_state=self.random_state)

        lanczos.fit(self.adjacency, 2)
        halko.fit(self.adjacency, 2)
        self.assertAlmostEqual(np.linalg.norm(lanczos.eigenvalues_ - halko.eigenvalues_), 0.)

        lanczos.fit(self.slr, 2)
        halko.fit(self.slr, 2)
        self.assertAlmostEqual(np.linalg.norm(lanczos.eigenvalues_ - halko.eigenvalues_), 0.)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for ranking API"""
import unittest

from sknetwork.data.test_graphs import test_bigraph, test_graph, test_digraph
from sknetwork.ranking import *


class TestPageRank(unittest.TestCase):

    def test_basic(self):
        methods = [PageRank(), Diffusion(), Closeness(), HITS(), Harmonic(), Katz(), Dirichlet()]
        for adjacency in [test_graph(), test_digraph()]:
            n = adjacency.shape[0]
            for method in methods:
                score = method.fit_transform(adjacency)
                self.assertEqual(score.shape, (n, ))
                self.assertTrue(score.min() >= 0)

    def test_bipartite(self):
        biadjacency = test_bigraph()
        n_row, n_col = biadjacency.shape

        methods = [BiPageRank(), BiDiffusion(), HITS(), BiKatz(), BiDirichlet()]
        for method in methods:
            method.fit(biadjacency)
            scores_row = method.scores_row_
            scores_col = method.scores_col_

            self.assertEqual(scores_row.shape, (n_row,))
            self.assertEqual(scores_col.shape, (n_col,))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for randomized matrix factorization"""

import unittest

from sknetwork.linalg.randomized_methods import *
from sknetwork.linalg.sparse_lowrank import SparseLR


class TestClusteringMetrics(unittest.TestCase):

    def test_dot(self):
        n = 5
        x = np.random.randn(n)
        mat = sparse.eye(n, format='csr')
        slr = SparseLR(mat, [(np.zeros(n), np.zeros(n))])

        y1 = safe_sparse_dot(x, slr)
        y2 = safe_sparse_dot(slr, x)
        self.assertAlmostEqual(np.linalg.norm(y1 - y2), 0)

        with self.assertRaises(NotImplementedError):
            safe_sparse_dot(slr, slr)

        y1 = safe_sparse_dot(slr, mat).dot(x)
        y2 = safe_sparse_dot(mat, slr).dot(x)
        self.assertAlmostEqual(np.linalg.norm(y1 - y2), 0)

    def test_range_finder(self):
        x = np.random.randn(3, 7)
        y1 = randomized_range_finder(x, size=2, n_iter=2, power_iteration_normalizer='auto')
        y2 = randomized_range_finder(x, size=2, n_iter=2, power_iteration_normalizer='LU')
        y3 = randomized_range_finder(x, size=2, n_iter=2, power_iteration_normalizer='QR')

        self.assertEqual(y1.shape, (3, 2))
        self.assertEqual(y2.shape, (3, 2))
        self.assertEqual(y3.shape, (3, 2))

    def test_eig(self):
        eigenvalues, eigenvectors = randomized_eig(sparse.identity(5, format='csr'), 2)
        self.assertTrue(np.allclose(eigenvalues, np.array([1., 1.])))
        self.assertTrue(eigenvectors.shape == (5, 2))

        eigenvalues, eigenvectors = randomized_eig(sparse.identity(5, format='csr'), 2, one_pass=True)
        self.assertTrue(np.allclose(eigenvalues, np.array([1., 1.])))
        self.assertTrue(eigenvectors.shape == (5, 2))

    def test_svd(self):
        self.sym_matrix = sparse.csr_matrix(np.array([[0, 1, 1, 1],
                                                      [1, 0, 0, 0],
                                                      [1, 0, 0, 1],
                                                      [1, 0, 1, 0]]))
        left_singular_vectors, singular_values, _ = randomized_svd(sparse.identity(5, format='csr'), 2)
        self.assertTrue(np.allclose(singular_values, np.array([1., 1.])))
        self.assertTrue(left_singular_vectors.shape == (5, 2))

        _, singular_values, _ = randomized_svd(self.sym_matrix, 2)
        self.assertTrue(np.allclose(singular_values, np.array([2.17, 1.48]), 1e-2))



"""linalg module"""
from sknetwork.linalg.auto_mode import auto_solver
from sknetwork.linalg.eig_solver import EigSolver, LanczosEig, HalkoEig
from sknetwork.linalg.normalization import diag_pinv, normalize
from sknetwork.linalg.operators import RegularizedAdjacency, LaplacianOperator, NormalizedAdjacencyOperator, \
    CoNeighborOperator
from sknetwork.linalg.polynome import Polynome
from sknetwork.linalg.randomized_methods import safe_sparse_dot, randomized_eig, randomized_svd
from sknetwork.linalg.sparse_lowrank import SparseLR
from sknetwork.linalg.svd_solver import SVDSolver, LanczosSVD, HalkoSVD



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 2019
@author: Thomas Bonald <bonald@enst.fr>
@author: Quentin Lutz <qlutz@enst.fr>
"""

import unittest

from sknetwork.data.test_graphs import test_graph
from sknetwork.hierarchy import Paris


class TestParis(unittest.TestCase):

    def test_options(self):
        paris = Paris(weights='uniform')
        adjacency = test_graph()
        dendrogram = paris.fit_transform(adjacency)
        n = adjacency.shape[0]
        self.assertEqual(dendrogram.shape, (n - 1, 4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for betweenness.py"""

import unittest
import numpy as np

from sknetwork.ranking.betweenness import Betweenness
from sknetwork.data.test_graphs import test_graph, test_graph_disconnect
from sknetwork.data.toy_graphs import bow_tie, star_wars


class TestBetweenness(unittest.TestCase):

    def test_basic(self):
        adjacency = test_graph()
        betweenness = Betweenness()
        scores = betweenness.fit_transform(adjacency)
        self.assertEqual(len(scores), adjacency.shape[0])

    def test_bowtie(self):
        adjacency = bow_tie()
        betweenness = Betweenness()
        scores = betweenness.fit_transform(adjacency)
        self.assertEqual(np.sum(scores > 0), 1)

    def test_disconnected(self):
        adjacency = test_graph_disconnect()
        betweenness = Betweenness()
        with self.assertRaises(ValueError):
            betweenness.fit(adjacency)

    def test_bipartite(self):
        adjacency = star_wars()
        betweenness = Betweenness()

        with self.assertRaises(ValueError):
            betweenness.fit_transform(adjacency)
